@@ -141,6 +141,8 @@ void QAbstractItemModelReplicaImplementation::initializeModelConnections()
     connect(this, &QAbstractItemModelReplicaImplementation::currentChanged, this, &QAbstractItemModelReplicaImplementation::onCurrentChanged);
     connect(this, &QAbstractItemModelReplicaImplementation::modelReset, this, &QAbstractItemModelReplicaImplementation::onModelReset);
     connect(this, &QAbstractItemModelReplicaImplementation::headerDataChanged, this, &QAbstractItemModelReplicaImplementation::onHeaderDataChanged);
+    connect(this, &QAbstractItemModelReplicaImplementation::layoutChanged, this, &QAbstractItemModelReplicaImplementation::onLayoutChanged);
+
 }
 
 inline void removeIndexFromRow(const QModelIndex &index, const QVector<int> &roles, CachedRowEntry *entry)
@@ -266,6 +268,8 @@ void QAbstractItemModelReplicaImplementation::onRowsInserted(const IndexList &pa
     auto parentItem = cacheData(parentIndex);
     q->beginInsertRows(parentIndex, start, end);
     parentItem->insertChildren(start, end);
+    for (int i = start; i <= end; ++i)
+        m_headerData[1].append(CacheEntry());
     q->endInsertRows();
     if (!parentItem->hasChildren && parentItem->columnCount > 0) {
         parentItem->hasChildren = true;
@@ -292,6 +296,8 @@ void QAbstractItemModelReplicaImplementation::onColumnsInserted(const IndexList
             return;
     q->beginInsertColumns(parentIndex, start, end);
     parentItem->columnCount += end - start + 1;
+    for (int i = start; i <= end; ++i)
+        m_headerData[0].append(CacheEntry());
     q->endInsertColumns();
     if (!parentItem->hasChildren && parentItem->children.size() > 0) {
         parentItem->hasChildren = true;
@@ -313,6 +319,7 @@ void QAbstractItemModelReplicaImplementation::onRowsRemoved(const IndexList &par
     q->beginRemoveRows(parentIndex, start, end);
     if (parentItem)
         parentItem->removeChildren(start, end);
+    m_headerData[1].erase(m_headerData[1].begin() + start, m_headerData[1].begin() + end + 1);
     q->endRemoveRows();
 }
 
@@ -671,7 +678,7 @@ void QAbstractItemModelReplicaImplementation::onHeaderDataChanged(Qt::Orientatio
     // TODO clean cache
     const int index = orientation == Qt::Horizontal ? 0 : 1;
     QVector<CacheEntry> &entries = m_headerData[index];
-    for (int i = first; i < last; ++i )
+    for (int i = first; i <= last && i < entries.size(); ++i )
         entries[i].data.clear();
     emit q->headerDataChanged(orientation, first, last);
 }
@@ -693,6 +700,41 @@ void QAbstractItemModelReplicaImplementation::fetchPendingHeaderData()
     m_pendingRequests.push_back(watcher);
 }
 
+void QAbstractItemModelReplicaImplementation::onLayoutChanged(const IndexList &parents,
+                                                              QAbstractItemModel::LayoutChangeHint hint)
+{
+    QList<QPersistentModelIndex> indexes;
+    for (const ModelIndex &parent : qAsConst(parents)) {
+        const QModelIndex parentIndex = toQModelIndex(IndexList{parent}, q);
+        indexes << QPersistentModelIndex(parentIndex);
+    }
+    QRemoteObjectPendingCallWatcher *watcher;
+    auto call = replicaCacheRequest(m_rootItem.children.cacheSize, m_initialFetchRolesHint);
+    watcher = new QRemoteObjectPendingCallWatcher(call);
+    m_pendingRequests.push_back(watcher);
+    connect(watcher, &QRemoteObjectPendingCallWatcher::finished, this, [this, watcher, indexes, hint]() {
+        Q_ASSERT(watcher->returnValue().canConvert<MetaAndDataEntries>());
+        const QSize size = watcher->returnValue().value<MetaAndDataEntries>().size;
+
+        q->layoutAboutToBeChanged(indexes, hint);
+        m_rootItem.clear();
+        if (size.height() > 0) {
+            m_rootItem.rowCount = size.height();
+            m_rootItem.hasChildren = true;
+        }
+
+        m_rootItem.columnCount = size.width();
+        if (m_initialAction == QtRemoteObjects::PrefetchData) {
+            auto entries = watcher->returnValue().value<MetaAndDataEntries>();
+            for (int i = 0; i < entries.data.size(); ++i)
+                fillCache(entries.data[i], entries.roles);
+        }
+        m_pendingRequests.removeAll(watcher);
+        watcher->deleteLater();
+        emit q->layoutChanged(indexes, hint);
+    });
+}
+
 static inline QVector<QPair<int, int> > listRanges(const QVector<int> &list)
 {
     QVector<QPair<int, int> > result;
